#!/usr/bin/python

import utmp
from UTMPCONST import *
import time, pwd, grp, os, string, sys, socket
from stat import *
from string import upper

def boguslist():
    # get list of (pid, tty, commandname)
    # bogus because we are just guessing by the output of ps
    pscmd = "ps ac" # seems to be common
    psoutput = os.popen(pscmd).readlines()
    output = []
    del psoutput[0]
    for i in psoutput:
        l = string.split(i)
        pid, tty, command = int(l[0]), l[1], l[-1]
        output.append( (pid, tty, command) )
    return output
    
def proctable():
    #make table of ttys and corresponding pids
    global proctbl
    proctbl = {}
    for i in boguslist():
        if not proctbl.has_key(i[1]):
            proctbl[i[1]] = []
        proctbl[i[1]].append( (i[0], i[2]) )
            

def header(now, boottime, ops, ver):
    print "Consortium for global VMSfication"
    print time.strftime("%A, %d-%b-%Y %H:%M,", now), "%s V%s" %(ops, ver)
    print "Cluster  contains 1 node, formed", \
          time.strftime("%A, %d-%b-%Y %H:%M",  time.localtime(boottime))
    print

def getrealname(gec):
    # get real name from gecos fiels
    return string.split(gec,",",1)[0]

def userlist(u, now, node, user=""):
    proctable()
    u.setutent()
    tnow = time.mktime(now)
    header = 0
    output = [] # list of output lines, without header
    while 1:
        b = u.getutent_dict()
        if not b:
            break
        if b['ut_type'] == USER_PROCESS:
            username = b['ut_user']
            if user and b['ut_user']<>user:
                continue
            try:
                pwnam = pwd.getpwnam(username)
            except KeyError:
                continue
            realname = getrealname(pwnam[4])
            username = upper(username) # yeah, right :-)
            if 0:
                try:
                    # this works on linux and freebsd
                    # unfortunately, even if it works,
                    # it does not work well because it always returns shell...
                    f = open("/proc/%i/cmdline" % b['ut_pid'], "r")
                    program = f.read()
                    program = os.path.split(program)[1]
                except:
                    program = "?"

            tty = b['ut_line']
            program = max(proctbl[tty])[1]
            program = string.split(program, ".")[0] # SunOS uses dots in names
                
            shell = os.path.split(pwnam[6])[1]
            if program == shell or (program[0]=="-" and program[1:]==shell):
                program = "$"
            t = time.localtime(b['ut_tv'][0])
            then = time.mktime(t)
            if tnow<then: # login in the future?
                login = "FUTURE"
            elif t[7] == now[7] and t[0] == now[0]: # today
                login = time.strftime("%H:%M", t)
            elif tnow-then<60*60*24*7: # this week
                login = time.strftime("%a", t)
            elif tnow-then<60*60*24*365.: # this year
                login = time.strftime("%d-%b", t)
            else: # way down in the past
                login = time.strftime("%Y", t)
            location = b['ut_host']
            if location:
                location = "Host: "+location
            else:
                location = b['ut_line']
            
            #length sanitation
            username = username[:12]
            realname = realname[:22]
            program = upper(program[:10])
            login = login[:6]
            node = node[:6]
            location = location[:20]
            
            if not header:
                print "%-12s%-23s%-10s%-7s%-7s%-20s" % \
                      ("Username","Real Name","Program","Login","Node","Location")
                header = 1
        
            output.append( "%-12s%-23s%-10s%-7s%-7s%-20s" % 
                   (username,realname,program,login,node,location) )
    output.sort()
    for i in output:
        print i
    return output

def lastlogin(u, user):
    lastlogin = 0
    u.setutent()
    while 1:
        b = u.getutent_dict()
        if not b:
            break
        if b['ut_type'] in (USER_PROCESS, DEAD_PROCESS) and \
           b['ut_user'] == user and \
           b['ut_tv'][0]>lastlogin:
            lastlogin = b['ut_tv'][0]

    u = utmp.UtmpRecord(WTMP_FILE)
    while 1:
        b = u.getutent_dict()
        if not b:
            break
        if b['ut_type'] in (USER_PROCESS, DEAD_PROCESS) and \
           b['ut_user'] == user and \
           b['ut_tv'][0]>lastlogin:
            lastlogin = b['ut_tv'][0]
    u.endutent()
    return lastlogin

def userplan(homedir):          
    try:
        f = open(homedir+"/.plan", "r")
        print "Plan:"
        while 1:
            l = f.readline()
            if not l:
                break
            print string.rstrip(l)
    except:
        pass
    

def oneuser(u, user):
    pwent = pwd.getpwnam(user)
    print "Login name: %-28sIn real life: %s" % \
         (upper(user), getrealname(pwent[4]))
    print " Directory: %-37sUIC: [%s,%s] ([%i,%i])" % \
          (pwent[5], 
           upper(grp.getgrgid(pwent[3])[0]), upper(user),
           pwent[3], pwent[2])
    l = lastlogin(u, user)
    if not l:
        print "Never logged in."
    else:
        print "Last login:", time.strftime("%A, %d-%b-%Y %H:%M", time.localtime(l))
    print
    userplan(pwent[5])

def guessbotttime(u):
    # try to find out boot time
    boottime = os.stat("/")[ST_CTIME] #fallback
    u.setutent()
    while 1:
        b = u.getutent_dict()
        if not b:
            break
        if b['ut_type'] == BOOT_TIME:
            boottime = b['ut_tv'][0]
    return boottime


if len(sys.argv) == 2 and "@" in sys.argv[1]: # remote
    user, host = string.split(sys.argv[1], "@", 1)
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    FINGER_PORT = 79
    s.connect( (host, FINGER_PORT) )
    s.send(user + '\r\n')
    while 1:
        buf = s.recv(1024)
        if not buf: break
        sys.stdout.write(buf)
    sys.stdout.flush()
    sys.exit(0)

un = os.uname()
ops = un[0]
node = un[1]
ver = un[2]
node = upper(node)
now = time.localtime(time.time())
a = utmp.UtmpRecord()
boottime = guessbotttime(a)

if len(sys.argv) == 1: # list of all local users
    header(now, boottime, ops, ver)
    r = userlist(a, now, node)
    if not r:
        print "No such processes."

else:
    #first find out if user exists
    user = sys.argv[1]
    try:
        pwd.getpwnam(user)
        r = userlist(a, now, node, user)
        if not r:
            print upper(user), "isn't logged in."
        print
        oneuser(a, user)
    except KeyError:
        lou = [] # list of matching users
        if len(user)>=3:
            for i in pwd.getpwall():
                rn = getrealname(i[4])
                if string.count(upper(rn), upper(user)):
                    lou.append( (i[0], rn) )
        if not lou:
            print upper(user), "does not match any user of this system."
        else:
            print 'Users who have "%s" in their real names:' % upper(user)
            for i in lou:
                print "%-13s- %s" % (upper(i[0]), i[1])
    
a.endutent()




#!/usr/bin/python

import utmp
from UTMPCONST import *
import time, pwd, grp, os, string, sys, socket, popen2
from stat import *
from string import lower


def getrealname(gec):
    # get real name from gecos fiels
    return string.split(gec,",",1)[0]

def formatidle(t):
    if t<30:
        return ""
    if t<80:
        r = "%ss" % int(t)
        return r
    if t<60*80:
        return "%sm" % int(t/60)
    if t<60*60*28:
        return "%sh" % int(t/60/60)
    if t<60*60*24*20:
        return "%sd" % int(t/60/60/24)
    return "DEAD"

def userlist(u, now, user=""):
    u.setutent()
    tnow = time.mktime(now)
    header = 0
    output = [] # list of output lines, without header
    while 1:
        b = u.getutent()
        if not b:
            break
        if b.ut_type == USER_PROCESS:
            username = b.ut_user
            if user and b.ut_user!=user:
                continue
            try:
                pwnam = pwd.getpwnam(username)
            except KeyError:
                pwnam = '?'

            tty = b.ut_line

            t = time.localtime(b.ut_tv[0])
            then = time.mktime(t)
            if tnow<then: # login in the future?
                login = "FUTURE"
            elif t[7] == now[7] and t[0] == now[0]: # today
                login = time.strftime("%H:%M", t)
            elif tnow-then<60*60*24*7: # this week
                login = time.strftime("%a", t)
            elif tnow-then<60*60*24*365.: # this year
                login = time.strftime("%d-%b", t)
            else: # way down in the past
                login = time.strftime("%Y", t)
            location = b.ut_host
            tty = b.ut_line
            try:
                s = os.stat("/dev/"+tty)
                p = s[ST_MODE] & 0x30 # 060 octal - test executable bit
                if tnow<s[ST_ATIME]:
                    idle = 0
                else:
                    idle = tnow-s[ST_ATIME]
                idle = formatidle(idle)
                if p:
                    p = ' '
                else:
                    p = '*'
            except:
                p = '?'
                
            if p == '?':
                continue
            #length sanitation
            username = username[:12]
            #realname = realname[:22]
            login = login[:6]
            location = location[:30]

            if not header:
                #print 60*"-"
                print ("%-12s%-7s%-4s%-2s%-8s%-30s" % \
                       ("USERNAME","Login","Idle","", "TTY","Location")
                      )
                #print 60*"-"
                header = 1
        
            output.append( "%-12s%-7s%4s%2s%-8s%-30s" % 
                   (username,login,idle,p,tty,location) )
    output.sort()
    for i in output:
        print(i)
    return output

def lastlogin(u, user):
    lastlogin = 0, ""
    u.setutent()
    while 1:
        b = u.getutent_dict()
        if not b:
            break
        if b.ut_type in (USER_PROCESS, DEAD_PROCESS) and \
           b.ut_user == user and \
           b.ut_tv[0]>lastlogin[0]:
            lastlogin = b.ut_tv[0], b.ut_host

    u = utmp.UtmpRecord(WTMP_FILE)
    while 1:
        b = u.getutent_dict()
        if not b:
            break
        if b.ut_type in (USER_PROCESS, DEAD_PROCESS) and \
           b.ut_user == user and \
           b.ut_tv[0]>lastlogin[0]:
            lastlogin = b.ut_tv[0], b.ut_host
    u.endutent()
    return lastlogin

def userplan(homedir):          
    try:
        f = open(homedir+"/.plan", "r")
        print("Plan:")
        while 1:
            l = f.readline()
            if not l:
                break
            print string.rstrip(l)
    except:
        pass


def oneuser(u, user):
    pwent = pwd.getpwnam(user)
    rn = getrealname(pwent[4])
    print ("Login name: %-30s In real life: %s" % (user, rn))
    print (" Directory: %-30s Shell: %s" % (pwent[5], pwent[6]))
    print ("            %-30s Group: [%s]" % ("", grp.getgrgid(pwent[3])[0]))
    l, h = lastlogin(u, user)
    if not l:
        print("Never logged in.")
    else:
        r = "Last login %-30s  " % time.strftime("%A, %d-%b-%Y %H:%M", time.localtime(l))
        if h:
            r = r+'from: '+h
        print(r)
    print('\n')
    userplan(pwent[5])

print('\n')

if len(sys.argv) == 2 and "@" in sys.argv[1]: # remote
    user, host = string.split(sys.argv[1], "@", 1)
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        FINGER_PORT = 79
        s.connect( (host, FINGER_PORT) )
        s.send(user + '\r\n')
        while 1:
            buf = s.recv(1024)
            if not buf: break
            sys.stdout.write(buf)
        sys.stdout.flush()
    except socket.error, why:
        print "ERROR:", why
    sys.exit(0)

now = time.localtime(time.time())
a = utmp.UtmpRecord()

if len(sys.argv) == 1: # list of all local users
    r = userlist(a, now)
    if not r:
        print "No such processes."

else:
    #first find out if user exists
    user = sys.argv[1]
    try:
        pwd.getpwnam(user)
        r = userlist(a, now, user)
        if not r:
            print '"%s" isn\'t logged in.' % user
        print
        oneuser(a, user)
    except KeyError:
        print '"%s" does not match any user of this system.' % user
a.endutent()




import utmpaccess
from UTMPCONST import *

import types

class UtmpEntry:
    def __init__(self, *arg, **dictarg):
        self.clear()
        if len(arg)==1:
            arg = arg[0]
        if (type(arg)==tuple or type(arg)==list) and arg:
            self.ut_type, self.ut_pid, self.ut_line, \
            self.ut_id, self.ut_user, self.ut_host, \
            self.ut_exit, self.ut_session, self.ut_tv, \
            self.ut_addr_v6 = arg
        elif type(arg)==dict:
            for i in arg.keys():
                self[i] = arg[i]
        elif type(arg)==types.InstanceType:
            self.ut_type, self.ut_pid, self.ut_line, \
            self.ut_id, self.ut_user, self.ut_host, \
            self.ut_exit, self.ut_session, self.ut_tv, \
            self.ut_addr_v6 = arg.ut_type, arg.ut_pid, arg.ut_line, \
            arg.ut_id, arg.ut_user, arg.ut_host, \
            arg.ut_exit, arg.ut_session, arg.ut_tv, \
            arg.ut_addr_v6
        for i in dictarg.keys():
            self[i] = dictarg[i]

    def clear(self):
            self.ut_type = EMPTY
            self.ut_pid = 0
            self.ut_line = ''
            self.ut_id = ''
            self.ut_user = ''
            self.ut_host = ''
            self.ut_exit = (0, 0)
            self.ut_session = 0
            self.ut_tv = (0, 0)
            self.ut_addr_v6 = (0, 0, 0, 0)

    def _as_tuple(self):
        return (self.ut_type, self.ut_pid, self.ut_line, \
            self.ut_id, self.ut_user, self.ut_host, \
            self.ut_exit, self.ut_session, self.ut_tv, \
            self.ut_addr_v6)

    def __getitem__(self, item):
        if item=='ut_type' or item==0:
            return self.ut_type
        elif item=='ut_pid' or item==1:
            return self.ut_pid
        elif item=='ut_line' or item==2:
            return self.ut_line
        elif item=='ut_id' or item==3:
            return self.ut_id
        elif item=='ut_user' or item==4:
            return self.ut_user
        elif item=='ut_host' or item==5:
            return self.ut_host
        elif item=='ut_exit' or item==6:
            return self.ut_exit
        elif item=='ut_session' or item==7:
            return self.ut_session
        elif item=='ut_tv' or item==8:
            return self.ut_tv
        elif item=='ut_addr_v6' or item==9:
            return self.ut_addr_v6
        else:
            raise IndexError("Bad key used to access UtmpEntry: "+repr(item))

    def __setitem__(self, item, val):
        if item=='ut_type' or item==0:
            self.ut_type=val
        elif item=='ut_pid' or item==1:
            self.ut_pid=val
        elif item=='ut_line' or item==2:
            self.ut_line=val
        elif item=='ut_id' or item==3:
            self.ut_id=val
        elif item=='ut_user' or item==4:
            self.ut_user=val
        elif item=='ut_host' or item==5:
            self.ut_host=val
        elif item=='ut_exit' or item==6:
            self.ut_exit=val
        elif item=='ut_session' or item==7:
            self.ut_session=val
        elif item=='ut_tv' or item==8:
            self.ut_tv=val
        elif item=='ut_addr_v6' or item==9:
            self.ut_addr_v6=val
        else:
            raise IndexError("Bad key used to access UtmpEntry: "+repr(item))

    def __repr__(self):
        fs = []
        for key in 'ut_type', 'ut_pid', 'ut_line', \
                   'ut_id', 'ut_user', 'ut_host', \
                   'ut_exit', 'ut_session', 'ut_tv', \
                   'ut_addr_v6':
            fs.append( key+'='+repr(self[key]) )
        fs = ', '.join(fs)
        r = fs
        r = 'utmp.UtmpEntry( %s )' % fs
        return r

class UtmpRecord:

    def __init__(self, fname=None):
        if fname:
            utmpaccess.utmpname(fname)
        self.fname = fname
        self.setutent()

    def _makeclass(self, a):
        if not a:
            return None
        return UtmpEntry(a)

    def setutent(self):
        "rewinds the file pointer to the beginning of the utmp file."
        utmpaccess.setutent()

    def endutent(self):
        "closes the utmp file."
        utmpaccess.endutent()

    def getutent(self):
        """reads a line from the current file position in the utmp file. It returns an
        UtmpEntry instance corresponding to a given line."""
        return self._makeclass(utmpaccess.getutent())

    def __iter__(self):
        return self

    def __next__(self):
        r = self.getutent()
        if not r:
            raise StopIteration
        return r

    next = __next__

    def pututline(self, *ut, **dictut):
        """writes the UtmpEntry provided as parameter into the utmp file. It uses getutid() to search
        for the proper place in the file to insert the new entry. If it cannot find an
        appropriate slot for ut, pututline() will append the new entry to the end of the file."""
        if len(ut) == 1:  # one tuple passed as argument
            u=ut[0]
        else:
            u = ut
        u = UtmpEntry(u, **dictut)
        utmpaccess.pututline(*u._as_tuple())

    def getutid(self, ut_type, ut_id=''):
        """searches forward from the current file position in the utmp
        file based upon ut_type. If ut_type is RUN_LVL,  BOOT_TIME, 
        NEW_TIME, or OLD_TIME, getutid() will find the first entry whose
        ut_type field matches ut_type argument. If ut_type is one of
        INIT_PROCESS, LOGIN_PROCESS, USER_PROCESS, or DEAD_PROCESS,
        getutid() will find the first entry whose ut_id field matches
        ut_id argument."""
        return self._makeclass(utmpaccess.getutid(ut_type, ut_id))

    def getutline(self, ut_line):
        """searches forward from the current file position in the
        utmp file. It scans entries whose ut_type is USER_PROCESS or
        LOGIN_PROCESS and returns the first one whose ut_line field matches 
        ut_lie argument."""
        return self._makeclass(utmpaccess.getutline(ut_line))

    getutent_dict = getutent
    pututline_dict = pututline
    getutid_dict = getutid
    getutline_dict = getutline

    def __del__(self):
        self.endutent()

    def __repr__(self):
        if not self.fname:
            fname = ''
        return 'utmp.UtmpRecord(%s)' % fname



#!/usr/bin/python
# poor man's who

import utmp
from UTMPCONST import *
import time

a = utmp.UtmpRecord()

print "%-10s %-10s %5s %-25s %-20s" % ("USER", "TTY", "PID", "HOST", "LOGIN")

for b in a: # example of using an iterator
    if b.ut_type == USER_PROCESS:
        print "%-10s %-10s %5i %-25s %-20s" % \
        (b.ut_user, b.ut_line, b.ut_pid, 
         b.ut_host, time.ctime(b.ut_tv[0]))
a.endutent()




#!/usr/bin/python
# scan utmp and remove bogus entries
# works only for systems that have ut_pid
# linux has, BSD has not

import utmp
from UTMPCONST import *
import time, os, string

a = utmp.UtmpRecord()

print ("Removing bogus entries:")
print ("%-10s %-10s %5s %-25s %-20s" % ("USER", "TTY", "PID", "HOST", "LOGIN"))

ps = os.popen("ps aux").readlines()[1:]
#ps = os.popen("ps -edf").readlines()[1:]
pids = {}
for i in ps:
    user, pid = i.split()[:2]
    pids[int(pid)] = user


for b in a:
    if b.ut_type == USER_PROCESS:
        if (b.ut_pid not in pids) or b.ut_user!=pids[b.ut_pid]:
            print ("%-10s %-10s %5i %-25s %-20s" % (b.ut_user, b.ut_line, b.ut_pid, b.ut_host, time.ctime(b.ut_tv[0])))
            b.ut_type = DEAD_PROCESS
            b.ut_host = ''
            b.ut_tv = (0, 0)
            a.pututline(b)
            a.getutent() # to move to next entry




#!/usr/bin/python
# poor man's last

import utmp
from UTMPCONST import *
import time

a = utmp.UtmpRecord(WTMP_FILE)

print ("%-10s %-10s %-30s %-20s" % ("USER", "TTY", "HOST", "LOGIN"))

while 1:
    b = a.getutent()
    if not b:
        break
    if b[0] == USER_PROCESS:
        print ("%-10s %-10s %-30s %-20s" % (b.ut_user, b.ut_line, b.ut_host, time.ctime(b.ut_tv[0])))
a.endutent()




#!/usr/bin/python
# poor man's last

import utmp
from UTMPCONST import *
import time

a = utmp.UtmpRecord(WTMP_FILE)

print "%-10s %-10s %-30s %-20s" % ("USER", "TTY", "HOST", "LOGIN")

while 1:
    b = a.getutent()
    if not b:
        break
    if b[0] == USER_PROCESS:
        print "%-10s %-10s %-30s %-20s" % (b[4], b[2], b[5], time.ctime(b[8][0]))
a.endutent()




#!/usr/bin/python
# scan utmp and remove bogus entries
# works only for systems that have ut_pid
# linux has, BSD has not

import utmp
from UTMPCONST import *
import time, os, string

a = utmp.UtmpRecord()

print "Removing bogus entries:"
print "%-10s %-10s %5s %-25s %-20s" % ("USER", "TTY", "PID", "HOST", "LOGIN")

ps = os.popen("ps aux").readlines()[1:]
#ps = os.popen("ps -edf").readlines()[1:]
pids = {}
for i in ps:
    user, pid = string.split(i)[:2]
    pids[int(pid)] = user

while 1:
    b = a.getutent()
    if not b:
        break
    if b[0] == USER_PROCESS:
        if (not pids.has_key(b[1])) or b[4]<>pids[b[1]]:
            print "%-10s %-10s %5i %-25s %-20s" % (b[4], b[2], b[1], b[5], time.ctime(b[8][0]))
            b = list(b)
            b[0] = DEAD_PROCESS
            b[4] = ''
            b[8] = (0, 0)
            a.pututline(b)
            a.getutent() # to move to next entry
a.endutent()




#!/usr/bin/python
# remove tty given as argument from utmp file

import utmp, sys
from UTMPCONST import *

a = utmp.UtmpRecord()

line = sys.argv[1]

b = a.getutline_dict(line)
if not b:
    print line, "is not occupied, cannot remove."
    sys.exit()
b['ut_type'] = DEAD_PROCESS
b['ut_line'] = ''
b['ut_user'] = ''
a.pututline_dict(b)
a.endutent()

